#include <stdio.h>
#include <string.h>
#include <app/Application.h>
#include <app/Roster.h>
#include <stdlib.h>
#include <unistd.h>

#include <interface/Alert.h>
#include <interface/Box.h>
#include <interface/Button.h>
#include <interface/Font.h>
#include <interface/Menu.h>
#include <interface/MenuItem.h>
#include <interface/MenuBar.h>
#include <interface/ScrollBar.h>
#include <interface/Screen.h>
#include <interface/ScrollBar.h>
#include <interface/ScrollView.h>
#include <interface/Input.h>
#include <interface/PopUpMenu.h>

#include <storage/File.h>
#include <storage/Path.h>     
#include <storage/NodeMonitor.h>     

#include "ColumnListView.h"
#include "CLVColumn.h"

#include "stringtokenizer/StringTokenizer.h"
#include "message/PortableMessage.h"
#include "reflector/StorageReflectConstants.h"
#include "reflector/StorageReflectUtils.h"
#include "netutil/NetworkUtilityFunctions.h"

#include "ShareWindow.h"
#include "ShareNetClient.h"
#include "ShareFileTransfer.h"
#include "ShareColumn.h"
#include "ShareMIMEInfo.h"
#include "RemoteUserItem.h"
#include "RemoteFileItem.h"

#define DEFAULT_SERVER "12.18.240.15"  // Ben's Linux box (email: creator@octavius.org)

// Window sizing constraints
#define MIN_WIDTH  600
#define MIN_HEIGHT 280
#define MAX_WIDTH  65535
#define MAX_HEIGHT 65535

// Default window position
#define WINDOW_START_X 30
#define WINDOW_START_Y 100          
#define WINDOW_START_W 640
#define WINDOW_START_H 400          

// Size constants for the non-resizable parts of the GUI
#define USER_LIST_WIDTH    125
#define STATUS_VIEW_WIDTH  450
#define UPPER_VIEW_HEIGHT  27
#define CHAT_VIEW_HEIGHT   125

#define TEXT_ENTRY_HEIGHT  20

#define SPECIAL_COLUMN_CHAR 0x01

#define FILE_NAME_COLUMN_NAME    "\0011File Name"
#define FILE_OWNER_COLUMN_NAME   "\0012Owner Name"
#define FILE_SESSION_COLUMN_NAME "\0013Owner ID"

#define SHARE_SETTINGS_FILE "/boot/home/config/settings/beshare_settings"  // hard coded path -> bad programmer!  no biscuit!
#define DEFAULT_COLUMN_WIDTH 40.0f

static uint8 DefaultData[256] =
{
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0xFE, 0xFE, 0x62, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0x00, 0xFE, 0xFE, 0xFE, 0xB0, 0x62, 0x62, 0x62, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF,
        0xFF, 0x00, 0xFE, 0xFE, 0xFE, 0xB0, 0xB0, 0x89, 0xB0, 0x62, 0x62, 0x89, 0xFE, 0xFE, 0x00, 0x00,
        0x00, 0xFE, 0xFE, 0xFD, 0xFE, 0xFE, 0xB0, 0xB0, 0xB0, 0x89, 0xB0, 0xFE, 0xFE, 0x00, 0x00, 0xB0,
        0xB0, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xB0, 0xB0, 0xFD, 0xFD, 0x29, 0x89, 0x00, 0x00,
        0xFF, 0xB0, 0x00, 0x00, 0x00, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x89, 0x00, 0xB0, 0xB0,
        0xFF, 0xFF, 0xB0, 0x00, 0x89, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0xB0, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0x62, 0x89, 0x00, 0xFA, 0x00, 0xB0, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0x00, 0x00, 0x00, 0xB0, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0xB0, 0xFF, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

ShareWindow :: ShareWindow()
  : BWindow(BRect(WINDOW_START_X,WINDOW_START_Y,WINDOW_START_X+WINDOW_START_W,WINDOW_START_Y+WINDOW_START_H),"BeShare",B_TITLED_WINDOW,0L), _netClient(NULL), _queryEnabled(false), _isConnected(false), _isConnecting(false), _activeAttribs(PortableStringHashFunc), _columns(CStringHashFunc), _attribMenuItems(CStringHashFunc), _users(CStringHashFunc), _mimeInfos(CStringHashFunc), _defaultBitmap(BRect(0,0,15,15),B_COLOR_8_BIT,DefaultData,false,false)
{
   const float vMargin = 5.0f;
   const float hMargin = 5.0f;

   _users.SetCompareFunction(CStringCompareFunc);
   _columns.SetCompareFunction(CStringCompareFunc);
   _attribMenuItems.SetCompareFunction(CStringCompareFunc);
   _mimeInfos.SetCompareFunction(CStringCompareFunc);

   BMessenger toMe(this);

   // Attempt to load settings...  (TODO:  use find_directory() instead of a hard-coded file path)
   BMessage settingsMsg;
   BFile settingsFile(SHARE_SETTINGS_FILE, B_READ_ONLY);
   if (settingsFile.InitCheck() == B_NO_ERROR) (void) settingsMsg.Unflatten(&settingsFile);

   // Recall saved window position
   BRect pos;
   if (settingsMsg.FindRect("windowpos", &pos) == B_NO_ERROR)
   {
      // check to make sure the window isn't opening off-screen
      BRect screenBounds;
      {
         BScreen s;
         screenBounds = s.Frame();
      }
      if (pos.left > screenBounds.Width()) pos.left = WINDOW_START_X;
      if (pos.top > screenBounds.Height()) pos.top = WINDOW_START_Y;

      MoveTo(pos.left, pos.top);
      ResizeTo(pos.Width(), pos.Height());
   }

   // Recall our "active columns" from the settings message
   BMessage columnsSubMessage;
   if (settingsMsg.FindMessage("columns", &columnsSubMessage) == B_NO_ERROR)
   {
      char * name;
      type_code type;
      int32 count;
      for (int32 i=0; (columnsSubMessage.GetInfo(B_FLOAT_TYPE, i, &name, &type, &count) == B_NO_ERROR); i++)
      {
         float width;
         if (columnsSubMessage.FindFloat(name, &width) == B_NO_ERROR) _activeAttribs.Put(name, width);
      }
   }
   else
   {
      // Put default columns here (size, name, owner, ?)
      _activeAttribs.Put(FILE_NAME_COLUMN_NAME, 330.0f);
      _activeAttribs.Put(FILE_OWNER_COLUMN_NAME, 100.0f);
      _activeAttribs.Put("beshare:File Size", 60.0f);
   }

   // The directory to download files to
   (void)GetAppSubdir("downloads", _downloadsDir, true);

   // Set up our share-file server thread.  This thread is responsible for
   // accepting incoming connections from other BeShare clients and notifying
   // us so that we can start the file transfer to them.
   BMessage newSessionMessage(SHAREWINDOW_COMMAND_NEW_PEER_SESSION); 
   _shareServer = new MessageTransceiverThread(toMe);
   const char * logMessage = NULL;
   if (_shareServer->StartServerThread(0, newSessionMessage) != B_NO_ERROR)
   {
      delete _shareServer;  
      _shareServer = NULL;
 
      // Can't call LogMessage() here, because the GUI isn't set up
      // yet.  So remember to do it later.
      logMessage = "Couldn't start file-sharing server thread!";
   }

   // set up the net client: this is what talks to the MUSCLE server for us
   (void)GetAppSubdir("shared", _shareDir, true);
   _netClient = new ShareNetClient(_shareDir, _shareServer ? _shareServer->GetPort() : -1);
   AddHandler(_netClient);

   SetSizeLimits(MIN_WIDTH, MAX_WIDTH, MIN_HEIGHT, MAX_HEIGHT);

   _menuBar = new BMenuBar(BRect(), "Menu Bar");

   BMenu * menu1 = new BMenu("File");
   menu1->AddItem(_connectMenuItem = new BMenuItem("Connect to <server>", new BMessage(SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER), 'C'));
   menu1->AddItem(_disconnectMenuItem = new BMenuItem("Disconnect",     new BMessage(SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER), 'D'));
   menu1->AddItem(new BSeparatorItem);
   menu1->AddItem(new BMenuItem("About BeShare", new BMessage(SHAREWINDOW_COMMAND_ABOUT), 'A'));
   menu1->AddItem(new BSeparatorItem);
   menu1->AddItem(new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED), 'Q'));
   _menuBar->AddItem(menu1);

   _attribMenu = new BMenu("Attributes");
   _menuBar->AddItem(_attribMenu);

   AddChild(_menuBar);                          

   BRect contentBounds = Bounds();
   contentBounds.top = _menuBar->Bounds().Height();

   // Create group/area views (top level stuff)
   BView * contentView = new BView(contentBounds, "ContentView", B_FOLLOW_ALL_SIDES, 0);
   contentView->SetViewColor(BeBackgroundGrey);
   AddChild(contentView);

   {
      // Fill out the upperLevel view
      BRect upperViewBounds(0, 0, contentBounds.Width() - hMargin, UPPER_VIEW_HEIGHT);
      BView * upperView = new BView(upperViewBounds, "UpperView", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 0);
      upperView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(upperView);

      {
         // Fill out the query view
         BRect queryViewBounds(0, 0, upperViewBounds.Width()-STATUS_VIEW_WIDTH, upperViewBounds.Height());
         BView * queryView = new BView(queryViewBounds, NULL, B_FOLLOW_ALL_SIDES, 0);
         queryView->SetViewColor(BeBackgroundGrey);
         upperView->AddChild(queryView);

         const char * startupQuery;
         if (settingsMsg.FindString("query", &startupQuery) != B_NO_ERROR) startupQuery = "*.mp3";
         _fileNameQueryEntry = new BTextControl(BRect(0,6,queryViewBounds.Width()-hMargin,TEXT_ENTRY_HEIGHT), NULL, "Query:", startupQuery, new BMessage(SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY), B_FOLLOW_ALL_SIDES);
         _fileNameQueryEntry->SetTarget(toMe);
         _fileNameQueryEntry->SetDivider(36.0f);
         queryView->AddChild(_fileNameQueryEntry);
      }

      {
         // Fill out the status view
         BRect statusViewBounds(upperViewBounds.Width()-STATUS_VIEW_WIDTH, 0, upperViewBounds.Width(), upperViewBounds.Height());
         BView * statusView = new BView(statusViewBounds, NULL, B_FOLLOW_RIGHT| B_FOLLOW_TOP_BOTTOM, 0);
         statusView->SetViewColor(BeBackgroundGrey);
         upperView->AddChild(statusView);

         _enableQueryButton = new BButton(BRect(0,3,75,TEXT_ENTRY_HEIGHT), NULL, "Start Query", new BMessage(SHAREWINDOW_COMMAND_ENABLE_QUERY));
         statusView->AddChild(_enableQueryButton);

         _disableQueryButton = new BButton(BRect(77,3,152,TEXT_ENTRY_HEIGHT), NULL, "Stop Query", new BMessage(SHAREWINDOW_COMMAND_DISABLE_QUERY));
         statusView->AddChild(_disableQueryButton);

         const char * sn;
         if (settingsMsg.FindString("server", &sn) != B_NO_ERROR) sn = DEFAULT_SERVER;
         _serverEntry = new BTextControl(BRect(154, 6, 275, statusViewBounds.Height()), NULL, "Server:", sn, new BMessage(SHAREWINDOW_COMMAND_USER_CHANGED_SERVER));
         _serverEntry->SetTarget(toMe);
         _serverEntry->SetDivider(40.0f);
         statusView->AddChild(_serverEntry);   

         const char * un;
         if (settingsMsg.FindString("username", &un) != B_NO_ERROR) un = "Anonymous";
         _netClient->SetLocalUserName(un);
         _userNameEntry = new BTextControl(BRect(279,6,STATUS_VIEW_WIDTH,statusViewBounds.Height()), NULL, "User Name:", un, new BMessage(SHAREWINDOW_COMMAND_USER_CHANGED_NAME));
         _userNameEntry->SetDivider(57.0f);
         _userNameEntry->SetTarget(toMe);
         
         statusView->AddChild(_userNameEntry);
      }
   }
    
   BRect resultsBounds(2, UPPER_VIEW_HEIGHT, contentBounds.Width()-(USER_LIST_WIDTH+hMargin), contentBounds.Height()-CHAT_VIEW_HEIGHT);
   {
      BView * resultsView = new BView(resultsBounds, "IOView", B_FOLLOW_ALL_SIDES, 0);
      resultsView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(resultsView);

      CLVContainerView* clvContainerView;
      _resultsView = new ColumnListView(BRect(hMargin, vMargin, resultsView->Bounds().Width()-(B_V_SCROLL_BAR_WIDTH+2), resultsView->Bounds().Height()-(vMargin+TEXT_ENTRY_HEIGHT+B_H_SCROLL_BAR_HEIGHT+8)),&clvContainerView,NULL,B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE,B_MULTIPLE_SELECTION_LIST,false,true,true,true,B_FANCY_BORDER);
      _resultsView->SetSortFunction((CLVCompareFuncPtr) CompareFunc);
      _resultsView->SetTarget(toMe);
      _resultsView->SetSelectionMessage(new BMessage(SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED));
      resultsView->AddChild(clvContainerView);

     _resultsView->AddColumn(new CLVColumn("", 20.0f, CLV_LOCK_AT_BEGINNING | CLV_NOT_MOVABLE | CLV_NOT_RESIZABLE));

     _requestDownloadsButton = new BButton(BRect(_resultsView->Bounds().left+2, resultsView->Bounds().Height()-(TEXT_ENTRY_HEIGHT+vMargin+1),resultsBounds.right-250,resultsBounds.Height()-(vMargin)), NULL, "Download Selected Files", new BMessage(SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS), B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
     resultsView->AddChild(_requestDownloadsButton);

     _clearFinishedDownloadsButton = new BButton(BRect(resultsBounds.right-245, resultsView->Bounds().Height()-(TEXT_ENTRY_HEIGHT+vMargin+1),resultsBounds.right-2,resultsBounds.Height()-(vMargin)), NULL, "Clear Finished/Failed Transfers", new BMessage(SHAREWINDOW_COMMAND_CLEAR_FINISHED_DOWNLOADS), B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM);
     resultsView->AddChild(_clearFinishedDownloadsButton);
   }

   { 
      BRect transferBounds(resultsBounds.right+hMargin, resultsBounds.top+vMargin-2, contentBounds.Width()-hMargin, resultsBounds.bottom-(vMargin-2));
      BView * transferView = new BView(transferBounds, NULL, B_FOLLOW_RIGHT | B_FOLLOW_TOP_BOTTOM, 0);
      transferView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(transferView);

      _transferList = new BListView(BRect(2, 2, transferBounds.Width()-(2+B_V_SCROLL_BAR_WIDTH), transferBounds.Height()-(5+TEXT_ENTRY_HEIGHT+vMargin)), NULL, B_MULTIPLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES); 
      _transferList->SetViewColor(BeBackgroundGrey);
      _transferList->SetTarget(toMe);
      _transferList->SetSelectionMessage(new BMessage(SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED));
      transferView->AddChild(new BScrollView(NULL, _transferList, B_FOLLOW_ALL_SIDES, 0L, false, true, B_FANCY_BORDER));

     _cancelTransfersButton = new BButton(BRect(0, _transferList->Frame().bottom+vMargin+2, transferBounds.Width(), transferBounds.Height()-1), NULL, "Remove Selected", new BMessage(SHAREWINDOW_COMMAND_CANCEL_DOWNLOADS), B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
     transferView->AddChild(_cancelTransfersButton);
   }

   BRect chatViewBounds(hMargin, contentBounds.Height()+vMargin-CHAT_VIEW_HEIGHT, contentBounds.Width()-(USER_LIST_WIDTH+hMargin), contentBounds.Height()-vMargin);
   {
      BView * chatView = new BView(chatViewBounds, NULL, B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM, 0); 
      chatView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(chatView);

      BRect chatTextBounds(0, 0, chatViewBounds.Width(), chatViewBounds.Height()-(TEXT_ENTRY_HEIGHT+vMargin));
      BBox * box = new BBox(chatTextBounds, NULL, B_FOLLOW_ALL_SIDES);
      chatView->AddChild(box);
      
      _chatText = new BTextView(BRect(2,2,chatTextBounds.Width()-(2+B_V_SCROLL_BAR_WIDTH), chatTextBounds.Height()-2), NULL, chatTextBounds, B_FOLLOW_ALL_SIDES);
      _chatText->MakeEditable(false);
      _chatText->SetStylable(true);
      box->AddChild(new BScrollView(NULL, _chatText, B_FOLLOW_ALL_SIDES, 0L, false, true, B_FANCY_BORDER));

      _textEntry = new BTextControl(BRect(0, chatViewBounds.Height()-TEXT_ENTRY_HEIGHT, chatViewBounds.Width(), chatViewBounds.Height()), NULL, "Chat:", NULL, new BMessage(SHAREWINDOW_COMMAND_SEND_CHAT_TEXT), B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
      _textEntry->SetDivider(27.0f);
      _textEntry->SetTarget(toMe);
      chatView->AddChild(_textEntry);
   }

   { 
      BView * userListView = new BView(BRect(chatViewBounds.right+hMargin, chatViewBounds.top, contentBounds.Width()-hMargin, chatViewBounds.bottom), "UserListView", B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM, 0);
      userListView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(userListView);

      CLVContainerView* clvContainerView;
      _usersView = new ColumnListView(BRect(2, 2, userListView->Bounds().Width()-(B_V_SCROLL_BAR_WIDTH+2), userListView->Bounds().Height()-2),&clvContainerView,NULL,B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE,B_SINGLE_SELECTION_LIST,false,false,true,true,B_FANCY_BORDER);
      _usersView->AddColumn(new CLVColumn("Name", _usersView->Bounds().Width()*.70f, CLV_SORT_KEYABLE|CLV_LOCK_AT_BEGINNING));
      _usersView->AddColumn(new CLVColumn("ID", _usersView->Bounds().Width()*.29f, CLV_SORT_KEYABLE|CLV_LOCK_AT_END));
      userListView->AddChild(clvContainerView);
   }

   UpdateConnectStatus(true);
   UpdateQueryEnabledStatus();

   // Columns that we know will available for any file we can put up right away
   CreateColumn(NULL, FILE_NAME_COLUMN_NAME, false);
   CreateColumn(NULL, "beshare:File Size", false);
   CreateColumn(NULL, "beshare:Kind", false);
   CreateColumn(NULL, FILE_OWNER_COLUMN_NAME, false);
   CreateColumn(NULL, FILE_SESSION_COLUMN_NAME, false);

   LogMessage(LOG_INFORMATION_MESSAGE, "BeShare ready.  Select Connect from the File menu to begin!");
   if (logMessage) LogMessage(LOG_ERROR_MESSAGE, logMessage);
}


status_t
ShareWindow :: GetAppSubdir(const char * subDirName, BDirectory & subDir, bool createIfNecessary) const
{
   app_info appInfo;
   be_app->GetAppInfo(&appInfo);
   BEntry appEntry(&appInfo.ref);
   appEntry.GetParent(&appEntry);  // get the directory it's in
   BPath path(&appEntry);
   BPath subPath(&appEntry);
   subPath.Append(subDirName);

   // If the directory is already there, use it
   if (subDir.SetTo(subPath.Path()) == B_NO_ERROR) return B_NO_ERROR;

   // Directory not there?  Shall we create it then?
   if (createIfNecessary)
   {
      BDirectory appDir(path.Path());
      if ((appDir.InitCheck() == B_NO_ERROR)&&(appDir.CreateDirectory(subDirName, &subDir) == B_NO_ERROR)) return B_NO_ERROR;
   }
   return B_ERROR;  // oops, couldn't get it
}

ShareWindow :: ~ShareWindow()
{
   // Save settings...
   BMessage settingsMsg;
   if (_netClient->GetLocalUserName()[0]) settingsMsg.AddString("username", _netClient->GetLocalUserName());
   if (_fileNameQueryEntry->Text()[0]) settingsMsg.AddString("query", _fileNameQueryEntry->Text());
   if (_connectedTo.Cstr()[0]) settingsMsg.AddString("server", _connectedTo.Cstr());
   
   BRect windowpos = Frame();
   settingsMsg.AddRect("windowpos", windowpos);

   BMessage columnsSubMessage;
   PortableHashtableIterator<PortableString, float> colIter = _activeAttribs.GetIterator();
   PortableString * nextString;
   while((nextString = colIter.GetNextKey()) != NULL) 
   {
      float width = *colIter.GetNextValue();  // default width in case we can't find the column itself for some reason
      float colWidth = 0.0f;

      ShareColumn * sc;
      if (_columns.Get(nextString->Cstr(), sc)) colWidth = sc->Width();
      columnsSubMessage.AddFloat(nextString->Cstr(), (colWidth > 0.0f) ? colWidth : width);
   }
   settingsMsg.AddMessage("columns", &columnsSubMessage);

   BFile settingsFile(SHARE_SETTINGS_FILE, B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE);
   if (settingsFile.InitCheck() == B_NO_ERROR) (void) settingsMsg.Flatten(&settingsFile);

   ClearUsers();

   delete _netClient;
   delete _shareServer;
}

void
ShareWindow ::
ClearUsers()
{
   _usersView->RemoveItems(0, _usersView->CountItems());
   PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
   RemoteUserItem * next;
   while(iter.GetNextValue(next)) delete next;
   _users.Clear();
}

// Hash function that just uses the pointer value as a hash value
static uint32 RemoteUserItemPointerHash(RemoteUserItem * const & ptr);
static uint32 RemoteUserItemPointerHash(RemoteUserItem * const & ptr)
{
   return (uint32)ptr;
}

void 
ShareWindow ::
RequestDownloads()
{
   // First, collate the files by remote user.  We do this so that we only have to
   // make one TCP connection to each remote host, and download all files that are
   // to come from him one at a time over that.  It's more efficient than connecting
   // to him several times in parallel.
   PortableHashtable<RemoteUserItem *, ShareFileTransfer *> newTransferSessions(RemoteUserItemPointerHash);
   int32 nextIndex;
   for (int32 i=0; ((nextIndex =_resultsView->CurrentSelection(i)) >= 0); i++)
   {
      RemoteFileItem * item = (RemoteFileItem *) _resultsView->ItemAt(nextIndex);
      RemoteUserItem * owner = item->GetOwner();
      if ((owner->GetHostName()[0])&&(owner->GetPort() > 0))
      {
         ShareFileTransfer * xfer;
         if (newTransferSessions.Get(owner, xfer) == false)
         {
            xfer = new ShareFileTransfer(_downloadsDir, _netClient->GetLocalSessionID());
            AddHandler(xfer);
            newTransferSessions.Put(owner, xfer);
         }
         xfer->AddRequestedFileName(item->GetFileName());
      }
      else 
      {
         PortableString errStr("Can't download from user ");
         errStr += owner->GetHandle();
         errStr += ", no connection information available.";
         LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
      }
   }

   // Set up the ShareFileTransfers to await their incoming connections from the remote owners...
   PortableHashtableIterator<RemoteUserItem *, ShareFileTransfer *> iter = newTransferSessions.GetIterator();
   ShareFileTransfer * nextXfer;
   while(iter.GetNextValue(nextXfer))
   {
      RemoteUserItem * nextUser = *(iter.GetNextKey());
      if (nextXfer->StartConnectDownloadSession(nextUser->GetHostName(), nextUser->GetPort(), nextUser->GetHandle()) == B_NO_ERROR)
      {
         _transferList->AddItem(nextXfer);
         _transferList->ScrollTo(0, _transferList->CountItems()*TRANSFER_ITEM_HEIGHT);  // scroll to the bottom
      }
      else
      {
         PortableString errStr("File download session to ");
         errStr += nextUser->GetHandle();
         errStr += "/";
         errStr += nextUser->GetSessionID();
         errStr += " failed to initialize!";
         LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
         RemoveHandler(nextXfer);
         delete nextXfer;  // he's toast
      }
   }
}

void ShareWindow :: MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case SHAREWINDOW_COMMAND_NEW_PEER_SESSION:
      {
         int socket;
         if ((_shareServer)&&(_shareServer->GetNextIncomingConnection(socket) == B_NO_ERROR))
         {
            ShareFileTransfer * newSession = new ShareFileTransfer(_shareDir, _netClient->GetLocalSessionID());
            AddHandler(newSession);
            if (newSession->StartUploadSession(socket) == B_NO_ERROR) _transferList->AddItem(newSession);
            else
            {
               LogMessage(LOG_ERROR_MESSAGE, "Couldn't start ShareFileTransfer session to serve remote transfer request!");
               RemoveHandler(newSession);
               delete newSession;
            }
            UpdateDownloadButtonStatus();
         }
      }
      break;

      case SHAREWINDOW_COMMAND_REMOVE_SESSION:
      {
         ShareFileTransfer * who;
         if (msg->FindPointer("who", (void **) &who) == B_NO_ERROR)
         {
            RemoveHandler(who);
            _transferList->RemoveItem(who);
            delete who;
         }
      }
      break;

      case SHAREWINDOW_COMMAND_CLEAR_FINISHED_DOWNLOADS:
      {
         for (int i=_transferList->CountItems()-1; i>=0; i--)
         {
            ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(i);
            if ((next->IsConnected() == false)&&(next->IsConnecting() == false))
            {
               RemoveHandler(next);
               _transferList->RemoveItem(i);
               delete next;  
            }
         }
         UpdateDownloadButtonStatus();
      }
      break;

      case SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED:
         UpdateDownloadButtonStatus();
      break;

      case SHAREWINDOW_COMMAND_CANCEL_DOWNLOADS:
      {
         PortableQueue<ShareFileTransfer *> killList;
         int32 nextIndex;
         for (int i=0; ((nextIndex = _transferList->CurrentSelection(i)) >= 0); i++) killList.AddTail((ShareFileTransfer *)_transferList->ItemAt(nextIndex));
         for (int j=killList.CountItems()-1; j>=0; j--)
         {
            RemoveHandler(killList[j]);
            _transferList->RemoveItem(killList[j]);
            delete killList[j];
         }
         UpdateDownloadButtonStatus();
      }
      break;

      case SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS:
         RequestDownloads();
         _resultsView->DeselectAll();
         UpdateDownloadButtonStatus();
      break;
 
      case SHAREWINDOW_COMMAND_USER_CHANGED_NAME:
      {
         _netClient->SetLocalUserName(_userNameEntry->Text());
         PortableString str("Your name has been changed to ");
         str += _netClient->GetLocalUserName();
         LogMessage(LOG_INFORMATION_MESSAGE, str.Cstr());
      }   
      break;

      case SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER:
         _netClient->DisconnectFromServer();
      break;

      case SHAREWINDOW_COMMAND_ABOUT:
         (new BAlert("About BeShare", "BeShare v1.10b\nby Jeremy Friesner\njaf@lcsaudio.com", "Hoopy"))->Go(NULL);
      break;

      case SHAREWINDOW_COMMAND_SEND_CHAT_TEXT:
      {
         PortableString text = _textEntry->Text();
         if (text.EndsWith("\n")) text = text.Substring(0, text.Length()-1);
         _netClient->SendChatMessage("*", text.Cstr());
         LogMessage(LOG_LOCAL_USER_CHAT_MESSAGE, text.Cstr());
         _textEntry->SetText("");
      }
      break;

      case SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER:
      {
         const char * server = _serverEntry->Text();
         if (server)
         {
            _connectedTo = server;  // save this for later, when we're connected
            StringTokenizer tok(server, " :");
            const char * host = tok.GetNextToken();
            if (host)
            {
               const char * portStr = tok.GetNextToken();
               int port = portStr ? atoi(portStr) : 0;
               if (port == 0) port = 2960;
               _netClient->ConnectToServer(host, port ? port : 2960);
            }
         }
         UpdateConnectStatus(true);
      }
      break;

      case SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY:
         if (_queryEnabled) 
         {
            SetQueryEnabled(false);
            SetQueryEnabled(true);
         }
      break;
 
      case SHAREWINDOW_COMMAND_USER_CHANGED_SERVER:
         UpdateConnectStatus(false);
      break;

      case SHAREWINDOW_COMMAND_ENABLE_QUERY:
      {
         SetQueryEnabled(true);
         PortableString str("Started query: [");
         str += _fileNameQueryEntry->Text(); 
         str += "]";
         LogMessage(LOG_INFORMATION_MESSAGE, str.Cstr());
      }
      break;

      case SHAREWINDOW_COMMAND_DISABLE_QUERY:
         SetQueryEnabled(false);
         LogMessage(LOG_INFORMATION_MESSAGE, "Stopped Query.");
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_COLUMN:
      {
         const char * attrName;
         if (msg->FindString("attrib", &attrName) == B_NO_ERROR)
         {
            BMenuItem * mi;
            if (_attribMenuItems.Get(attrName, mi))
            {
               mi->SetMarked(!mi->IsMarked());
               ShareColumn * sc;
               if (_columns.Get(attrName, sc))
               {
                  bool isVisible = (_resultsView->IndexOfColumn(sc) >= 0);
                  if ((isVisible)&&(mi->IsMarked() == false)) 
                  {
                     _activeAttribs.Remove(sc->GetAttributeName());
                     _resultsView->RemoveColumn(sc);
                  }
                  else if ((isVisible == false)&&(mi->IsMarked())) 
                  {
                     float w = DEFAULT_COLUMN_WIDTH;
                     if (_activeAttribs.Get(sc->GetAttributeName(), w)) sc->SetWidth(w);
                                                                   else _activeAttribs.Put(sc->GetAttributeName(), sc->Width());
                     _resultsView->AddColumn(sc);
                  }
               }
            }
         }
      }
      break;

      default:
         BWindow :: MessageReceived(msg);
      break;
   }
}

void
ShareWindow ::
LogMessage(LogMessageType type, const char * text, const char * optSessionID)
{
   const char * preamble = "???"; 
   const rgb_color black = {0, 0, 0, 255};
   rgb_color color = black;  // default
   PortableString temp;

   switch(type)
   {
      case LOG_INFORMATION_MESSAGE:
         preamble = "System";
         color.blue = 128;
      break;

      case LOG_WARNING_MESSAGE:
         preamble = "Warning";
         color.blue = color.green = 128;
      break;

      case LOG_ERROR_MESSAGE:
         preamble = "Error";
         color.red = 128;
      break;

      case LOG_LOCAL_USER_CHAT_MESSAGE:
         temp = "(";
         temp += _netClient->GetLocalSessionID();
         temp += ") ";
         temp += _netClient->GetLocalUserName();
         preamble = temp.Cstr();
         color.green = 128;
      break;

      case LOG_REMOTE_USER_CHAT_MESSAGE:
         if (optSessionID)
         {
            temp = "(";
            temp += optSessionID;
            temp += ") ";
            temp += GetUserNameForSession(optSessionID);
            preamble = temp.Cstr();
            color.green = 1;  // if I don't put this, the text doesn't show up???
         }
      break;
   }

   text_run_array style;
   style.count = 1;
   style.runs[0].offset = 0;
   style.runs[0].font = *be_bold_font;
   style.runs[0].color = color;

   _chatText->Insert(_chatText->TextLength(), preamble, strlen(preamble), &style);

   // now back to black
   style.runs[0].color = black;
   style.runs[0].font = *be_plain_font;

   _chatText->Insert(_chatText->TextLength(), ": ", 2, &style);
   _chatText->Insert(_chatText->TextLength(), text, strlen(text));
   _chatText->Insert(_chatText->TextLength(), "\n", 2);
   _chatText->ScrollToOffset(_chatText->TextLength()-2);
}

 
const char *
ShareWindow ::
GetUserNameForSession(const char * sessionID) const
{
   RemoteUserItem * user;
   return (_users.Get(sessionID, user)) ? user->GetHandle() : "<unknown>";
}

void
ShareWindow ::
SetQueryEnabled(bool e)
{
   if (e != _queryEnabled)
   {
      _queryEnabled = e;
      if (_queryEnabled) 
      {
         // If there are regexp chars in the name, take it verbatim;
         // otherwise add *'s around the edges to make it a substring search.
         PortableString exp(_fileNameQueryEntry->Text());
         const char regexpchars[] = "^$+?*[].\\{}";
         bool foundRegExp = false;
         for (int i=0; i<sizeof(regexpchars)-1; i++)
         {
            if (strchr(exp.Cstr(), regexpchars[i]))
            {
               foundRegExp = true;
               break;
            }
         }
         if (foundRegExp == false)
         {
            PortableString str("*");
            str += exp;
            str += '*';
            exp = str;
         }

         // Make the query case-insensitive...
         PortableString noCaseExp;
         for (int i=0; i<exp.Length(); i++)
         {
            char next = exp[i];
            if ((next >= 'A')&&(next <= 'Z')) 
            { 
               char buf[5];
               sprintf(buf, "[%c%c]", next, next+('a'-'A'));
               noCaseExp += buf;
            }
            else if ((next >= 'a')&&(next <= 'z'))
            {
               char buf[5];
               sprintf(buf, "[%c%c]", next, next+('A'-'a'));
               noCaseExp += buf;
            }
            else if (next == '/') 
            {
               noCaseExp += '?';  // can't have slashes in our filenames!
            }
            else noCaseExp += next;
         }

         ClearResults();
         _netClient->StartQuery("*", noCaseExp.Cstr());
      }
      else _netClient->StopQuery();

      UpdateQueryEnabledStatus();
   }
}

void
ShareWindow ::
UpdateQueryEnabledStatus()
{
   _enableQueryButton->SetEnabled((_isConnected)&(!_queryEnabled));
   _disableQueryButton->SetEnabled((_isConnected)&(_queryEnabled));
}

void
ShareWindow ::
UpdateConnectStatus(bool titleToo)
{
   const char * sname = _serverEntry->Text();

   bool c = ((_isConnected)||(_isConnecting));
   _connectMenuItem->SetEnabled(!c);
   if (sname[0] == '\0') sname = "???";
   _disconnectMenuItem->SetEnabled(c);

   char buf[200];
   if (titleToo)
   {
      strcpy(buf, "BeShare");
      if ((_isConnecting)||(_isConnected))
      {
         strcat(buf, _isConnected ? " - Connected to " : " - Connecting to ");
         strncat(buf, _connectedTo.Cstr(), sizeof(buf));
      }
      buf[sizeof(buf)-1] = '\0';
      SetTitle(buf);
   }

   strcpy(buf, "Connect to ");
   strncat(buf, sname, sizeof(buf));
   buf[sizeof(buf)-1] = '\0';
   _connectMenuItem->SetLabel(buf);

   UpdateDownloadButtonStatus();
}

void
ShareWindow ::
UpdateDownloadButtonStatus()
{
   _requestDownloadsButton->SetEnabled((_isConnected)&&(_resultsView->CurrentSelection() >= 0));

   bool deadTransfersPresent = false;
   for (int i=_transferList->CountItems()-1; i>=0; i--)
   {
      ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(i);
      if ((next->IsConnected() == false)&&(next->IsConnecting() == false))
      {
         deadTransfersPresent = true;
         break;
      }
   }
   _clearFinishedDownloadsButton->SetEnabled(deadTransfersPresent);
   _cancelTransfersButton->SetEnabled(_transferList->CurrentSelection() >= 0);
}

bool
ShareWindow ::
QuitRequested()
{
   be_app->PostMessage(B_QUIT_REQUESTED);
   return true;
}


// The methods below are called by our ShareNetClient at the appropriate times.
void 
ShareWindow ::
SetConnectStatus(bool isConnecting, bool isConnected)
{
        if ((!_isConnected)&&(isConnected)) LogMessage(LOG_INFORMATION_MESSAGE, "Connection established.");
   else if ((!_isConnecting)&&(isConnecting)) LogMessage(LOG_INFORMATION_MESSAGE, "Connecting to server...");
   else if ((isConnecting == false)&&(isConnected == false))
   {
           if (_isConnected) LogMessage(LOG_ERROR_MESSAGE, "You are no longer connected to the MUSCLE server.");
      else if (_isConnecting) LogMessage(LOG_ERROR_MESSAGE, "Connection to server failed.");
   }

   _isConnecting = isConnecting;
   _isConnected  = isConnected;

   // If we're not connected anymore, make sure the display is clear
   if (_isConnected == false) 
   {
      ClearUsers();
      SetQueryEnabled(false);
   }

   UpdateConnectStatus(true);
   UpdateQueryEnabledStatus();
}


void 
ShareWindow ::
PutUser(const char * sessionID, const char * userName, const char * hostName, int port)
{
   RemoteUserItem * user;
   if (_users.Get(sessionID, user) == false)
   {
      user = new RemoteUserItem(this, sessionID);
      _users.Put(user->GetSessionID(), user);
      _usersView->AddItem(user);
      _usersView->SortItems();
   }
   if (userName) user->SetHandle(userName);
   if (hostName) user->SetHostName(hostName);
   if (port >= 0) user->SetPort(port);
}


void 
ShareWindow ::
RemoveUser(const char * sessionID)
{
   RemoteUserItem * user;
   if (_users.Remove(sessionID, user)) 
   {
      _usersView->RemoveItem(user);
      delete user;
   } 
}

void 
ShareWindow ::
PutResult(const char * sessionID, const char * fileName, PortableMessageRef fileInfo)
{
   PutUser(sessionID, NULL, NULL, -1);  // make sure the RemoteUserItem is present!
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->PutFile(fileName, fileInfo);
}

void 
ShareWindow ::
RemoveResult(const char * sessionID, const char * fileName)
{
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->RemoveFile(fileName);
}

void 
ShareWindow ::
ClearResults()
{
   PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
   RemoteUserItem * next;
   while(iter.GetNextValue(next)) next->ClearFiles();
}

void
ShareWindow ::
FileTransferConnected(ShareFileTransfer * who)
{
   RefreshTransferItem(who);
   UpdateDownloadButtonStatus();
}


void
ShareWindow ::
FileTransferDisconnected(ShareFileTransfer * who)
{
   // Uploads we just immediately get rid of, downloads
   // stick around till the user removes them.
   if (who->IsUploadSession())
   {
      // Can't just delete it here because it is calling
      // me!  So I'll send myself a message to do it later
      BMessage msg(SHAREWINDOW_COMMAND_REMOVE_SESSION);
      msg.AddPointer("who", who);
      PostMessage(&msg);
   }
   else RefreshTransferItem(who);

   UpdateDownloadButtonStatus();
}


const char * 
ShareWindow ::
GetFileCellText(const RemoteFileItem * item, int32 columnIndex) const
{
   return ((ShareColumn *)_resultsView->ColumnAt(columnIndex))->GetFileCellText(item);
}

const BBitmap *
ShareWindow ::
GetBitmap(const RemoteFileItem * item, int32 columnIndex) const
{
   const BBitmap * bmp = NULL;

   // For now, the only bitmap supported is the MIME type icon
   const char * mimeString;
   if (item->GetAttributes().FindString("beshare:Kind", &mimeString) == B_NO_ERROR)
   {
      ShareMIMEInfo * mi;
      if (_mimeInfos.Get(mimeString, mi)) bmp = mi->GetIcon();
   }
   return bmp ? bmp : &_defaultBitmap;
}

const BBitmap *
ShareWindow ::
GetBitmap(const char * mimeString)
{
   ShareMIMEInfo * mi = mimeString ? CacheMIMETypeInfo(mimeString) : NULL;
   const BBitmap * bmp = mi ? mi->GetIcon() : NULL;
   return bmp ? bmp : &_defaultBitmap;
}

ShareMIMEInfo *
ShareWindow ::
CacheMIMETypeInfo(const char * mimeString)
{
   ShareMIMEInfo * ret;
   if (_mimeInfos.Get(mimeString, ret) == false)
   {
      const char * label = mimeString;
      char buf[B_MIME_TYPE_LENGTH];
      BMimeType mt(mimeString);
      if ((mt.InitCheck()==B_NO_ERROR)&&(mt.GetShortDescription(buf) == B_NO_ERROR)) label = buf;

      ShareMIMEInfo * newInfo = new ShareMIMEInfo(label, mimeString);
      _mimeInfos.Put(mimeString, newInfo);
      _attribMenu->AddItem(newInfo);
      ret = newInfo;
   }
   return ret;
}

void
ShareWindow :: 
AddFileItem(RemoteFileItem * item)
{
   _resultsView->AddItem(item);

   const PortableMessage & attrs = item->GetAttributes();
   ShareMIMEInfo * optMimeInfo = NULL;
   const char * mimeString = NULL;
   if (attrs.FindString("beshare:Kind", &mimeString) == B_NO_ERROR) optMimeInfo = CacheMIMETypeInfo(mimeString);

   PortableMessageFieldNameIterator iter = attrs.GetFieldNameIterator();
   const char * next;
   while((next = iter.GetNextFieldName()) != NULL) CreateColumn(optMimeInfo, next, true);
   _resultsView->SortItems();
}



void
ShareWindow ::
CreateColumn(ShareMIMEInfo * optMimeInfo, const char * attrName, bool remote)
{
   if (_columns.ContainsKey(attrName) == false) 
   {
      const char * label = attrName;
      int type = ShareColumn::ATTR_MISC;

      if ((remote == false)&&(attrName[0] == SPECIAL_COLUMN_CHAR))
      {
         type = attrName[1]-'0';
         label += 2;
      }
      else
      {
         if (strncmp(attrName, "beshare:", 8) == 0)
         {
            // A BeShare var; handle these a bit differently  
            label += 8;
            optMimeInfo = NULL;   // these vars are not type specific
         }
         else
         {
            // It's a genuine attribute; try to find a better name for it
            if (optMimeInfo) 
            {
               const char * desc = optMimeInfo->GetAttributeDescription(attrName);
               if (desc) label = desc;
            }
         }
      }

      ShareColumn * column = new ShareColumn(type, attrName, label, 80.0f);
      _columns.Put(column->GetAttributeName(), column);

      BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_TOGGLE_COLUMN);
      msg->AddString("attrib", column->GetAttributeName());
      BMenuItem * mi = new BMenuItem(column->GetLabel(), msg);

      if (optMimeInfo) optMimeInfo->AddItem(mi);
                 else  _attribMenu->AddItem(mi);

      _attribMenuItems.Put(column->GetAttributeName(), mi);
      if (_activeAttribs.ContainsKey(attrName)) PostMessage(msg);
   }
}


void
ShareWindow :: RemoveFileItem(RemoteFileItem * item)
{
   _resultsView->RemoveItem(item);
}

void
ShareWindow ::
RefreshTransferItem(ShareFileTransfer * item)
{
   _transferList->InvalidateItem(_transferList->IndexOf(item));
}


void
ShareWindow ::
RefreshFileItem(RemoteFileItem * item)
{
   _resultsView->InvalidateItem(_resultsView->IndexOf(item));
}

void
ShareWindow ::
RefreshUserItem(RemoteUserItem * item)
{
   _usersView->InvalidateItem(_usersView->IndexOf(item));
}


int
ShareWindow ::
CompareFunc(const CLVListItem* item1, const CLVListItem* item2, int32 sortKey)
{
   return ((RemoteFileItem *)item1)->Compare(((RemoteFileItem *)item2), sortKey);  
}

int 
ShareWindow ::
Compare(const RemoteFileItem * rf1, const RemoteFileItem * rf2, int32 sortKey) const
{
   return ((ShareColumn *)_resultsView->ColumnAt(sortKey))->Compare(rf1, rf2);
}

